import os
import pkg_resources
import numpy as np
import pandas as pd 
from transposonmapper.utils import chromosomename_roman_to_arabic
from transposonmapper.importing import load_default_files,load_sgd_tab
from transposonmapper.properties import list_gene_names , gene_position, gene_aliases,chromosome_position
from transposonmapper.processing import chromosome_name_wigfile
from transposonmapper.processing.read_sgdfeatures import sgd_features   


def input_region(region,verbose):
    """Defines the region of interest for further processing 

    Parameters
    ----------
    region : str, int or list
        Enter chromosome as a number (or roman numeral) between 1 and 16 (I and XVI), 
        a list in the form ['chromosome number, start_position, end_position'] or a valid gene name.
    verbose : bool
        To allow warning messages. 
    

    Returns
    -------
    roi_start : NoneType, int 
        Describe the start of the genomic location if region=gene name , otherwise is a NoneType
    roi_end : NoneType, int 
        Describe the  end of the genomic location if region=gene name , otherwise is a NoneType

    region_type: str
        It is either "Gene" or "Chromosome" depending on the region provided
    chrom: str
        It is the name of the chromosome of the gene of interest if a gene name is provided as the region, otherwise
        is the roman description of the chromosome of interest. 
    """
    
    gff_file,_,gene_information_file=load_default_files()

    warningmessage = "WARNING: Specified chromosome or gene name not found. Enter chromosome as a number (or roman numeral) between 1 and 16 (I and XVI), a list in the form ['chromosome number, start_position, end_position'] or a valid gene name."

    
    if verbose == True:
        print('Selected region: ', region)

    if type(region) == str:
        if region.upper() in chromosomename_roman_to_arabic()[1]:
            chrom = region.upper()
            roi_start = None
            roi_end = None
            region_type = 'Chromosome'

        elif region.upper() in list_gene_names(gene_information_file):
            gene_pos_dict = gene_position(gff_file)
            region = region.upper()
            region_type='Gene'
            if region in gene_pos_dict:
                region_pos = gene_pos_dict.get(region)
                chrom = region_pos[0]
                roi_start = int(region_pos[1])
                roi_end = int(region_pos[2])
            else:
                gene_alias_dict = gene_aliases(gene_information_file)[0]
                region_alias = [key for key, val in gene_alias_dict.items() if region in val]
                if not region_alias == [] and region_alias[0] in gene_pos_dict:
                    region_pos = gene_pos_dict.get(region_alias[0])
                    chrom = region_pos[0]
                    roi_start = int(region_pos[1])-100
                    roi_end = int(region_pos[2])+100
                    
                else:
                    print(warningmessage)
                    return()
            
            

        else:
            print(warningmessage)
            return()


    elif type(region) == list:
        if type(region[0]) == str:
            chrom = region[0].upper()
        elif type(region[0]) == int:
            if region[0] in chromosomename_roman_to_arabic()[0]:
                chrom = chromosomename_roman_to_arabic()[0].get(region[0])
        else:
            print(warningmessage)
            return()
        roi_start = region[1]
        roi_end = region[2]
        region_type = 'Chromosome'


    elif type(region) == int:
        if region in chromosomename_roman_to_arabic()[0]:
            chrom = chromosomename_roman_to_arabic()[0].get(region)
            roi_start = None
            roi_end = None
        else:
            print(warningmessage)
            return()
        region_type = 'Chromosome'


    else:
        print(warningmessage)
        return()
    
    return roi_start,roi_end,region_type,chrom




def read_wig_file(wig_file,chrom):
    """Extract the information in the wigfile related to the chromosome of interested

    Parameters
    ----------
    wig_file : str
        absolute path of the wigfile location
    chrom : str
        Name of the chromosome in roman where to extract the informatiion from the wigfile

    Returns
    -------
    insrt_in_chrom_list : list 
        Genomic locations of transposon insertions in the given chromosome.
    reads_in_chrom_list: list
        How many reads are in each of the genomic locations of the insertions. 
    
    """

    with open(wig_file, 'r') as f:
        lines = f.readlines()

    chrom_start_line_dict, chrom_end_line_dict = chromosome_name_wigfile(lines)[1:]

    insrt_in_chrom_list = []
    reads_in_chrom_list = []
    for l in lines[chrom_start_line_dict.get(chrom):chrom_end_line_dict.get(chrom)]:
        insrt_in_chrom_list.append(int(l.strip('\n').split(' ')[0]))
        reads_in_chrom_list.append(int(l.strip('\n').split(' ')[1]))
        
    return insrt_in_chrom_list,reads_in_chrom_list


def read_pergene_file(pergene_insertions_file,chrom):
    """Reading the pergene file , the information per gene , related to where it starts and ends in the genome. 

    Parameters
    ----------
    pergene_insertions_file : str
        absolute path of the per gene file location
    chrom : str
        Name of the chromosome in roman where to extract the informatiion from the wigfile

    Returns
    -------
    gene_position_dict : dict 
        A dictionary describing the chromosome, start, and end location of every gene in the chromosome of interest. 
    """

    with open(pergene_insertions_file) as f:
        lines = f.readlines()


    gene_position_dict = {}
    for line in lines[1:]:
        line_split = line.strip('\n').split('\t')


        if line_split[1] == chrom:
            genename = line_split[0]
            gene_chrom = line_split[1]
            gene_start = int(line_split[2])
            gene_end = int(line_split[3])

            gene_position_dict[genename] = [gene_chrom, gene_start, gene_end] #DICT CONTAINING ALL GENES WITHIN THE DEFINED CHROMOSOME INCLUDING ITS START AND END POSITION


            geneinserts_str = line_split[4].strip('[]')
            if not geneinserts_str == '':
                geneinserts_list = [int(ins) for ins in geneinserts_str.split(',')]
            else:
                geneinserts_list = []


            genereads_str = line_split[5].strip('[]')
            if not genereads_str == '':
                genereads_list = [int(read) for read in genereads_str.split(',')]
            else:
                genereads_list = []


            if len(geneinserts_list) != len(genereads_list):
                print('WARNING: %s has different number of reads compared with the number of inserts' % genename )

    return gene_position_dict


def gene_location(chrom,gene_position_dict,verbose):
    """It gives structured information from the genes inside the chromosome of interest

    Parameters
    ----------
    chrom : str
        Name of the chromosome in roman where to extract the information.
    gene_position_dict : dict
        Dictionary with info of the genes inside the chromosome . It is the output of the function
        read_pergene_file

    verbose : bool 
        Same as main function dna_features. If True allows for warning messages. 

    Returns
    -------
    dna_dict: dict
        Dictionary with info about genes encoded in the sgd features file 
    start_ch : int
        Integer indicating the genomic location of where the chromosome of interest starts
    end_chr: int 
        Integer indicating the genomic location of where the chromosome of interest ends
    len_chr: int
        Length of the chromosome of interest
    feature_orf_dict: dict
        Dictionary with info about genes in the chromosome of interest 
    """

    gff_file,_,gene_information_file=load_default_files()
    sgd_features_file=load_sgd_tab()
    

    len_chr = chromosome_position(gff_file)[0].get(chrom)
    start_chr = chromosome_position(gff_file)[1].get(chrom)
    end_chr = chromosome_position(gff_file)[2].get(chrom)
    if verbose == True:
        print('Chromosome length = ', len_chr)

    dna_dict = {} #for each bp in chromosome, determine whether it belongs to a noncoding or coding region
    for bp in range(start_chr, end_chr + 1): #initialize dna_dict with all basepair positions as ['noncoding', None]
        dna_dict[bp] = ['noncoding', None] #form is: ['element_name', 'type']


    feature_orf_dict = sgd_features(sgd_features_file)[1]
    gene_alias_dict = gene_aliases(gene_information_file)[0]


    for gene in gene_position_dict:
        if gene in feature_orf_dict:
            if (not gene.endswith("-A") and not feature_orf_dict.get(gene)[1] == 'Verified') and (not gene.endswith("-B") and not feature_orf_dict.get(gene)[1] == 'Verified'):
                for bp in range(gene_position_dict.get(gene)[1]+start_chr, gene_position_dict.get(gene)[2]+start_chr+1):
                    dna_dict[bp] = [gene, "Gene; "+feature_orf_dict.get(gene)[1]]
        else:
            gene_alias = [key for key, val in gene_alias_dict.items() if gene in val][0]
            for bp in range(gene_position_dict.get(gene)[1]+start_chr, gene_position_dict.get(gene)[2]+start_chr+1):
                dna_dict[bp] = [gene_alias, "Gene; "+feature_orf_dict.get(gene_alias)[1]]

    return dna_dict,start_chr,end_chr,len_chr,feature_orf_dict


def feature_position(feature_dict, chrom, start_chr, dna_dict, feature_type=None):
    """ Get features for every gene in the chromosome of interest 

    Parameters
    ----------
    feature_dict : dict
        output of sgd_features(sgd_features_file)[i]
    chrom : str
        Name of the chromosome in roman where to extract the information.
    start_chr : int
        [description]
    dna_dict : dict 
        first output of the gene_location function 
    feature_type : [type], optional
        [description], by default None

    Returns
    -------
     dict 

    """
    
    position_dict = {}
    for feat in feature_dict:
        if feature_dict.get(feat)[5] == chrom:
#            if feat.startswith("TEL") and feat.endswith('L'): #correct for the fact that telomeres at the end of a chromosome are stored in the reverse order.
            if int(feature_dict.get(feat)[6]) > int(feature_dict.get(feat)[7]):
                position_dict[feat] = [feature_dict.get(feat)[5], feature_dict.get(feat)[7], feature_dict.get(feat)[6]]
            else:
                position_dict[feat] = [feature_dict.get(feat)[5], feature_dict.get(feat)[6], feature_dict.get(feat)[7]]


    for feat in position_dict:
        for bp in range(int(position_dict.get(feat)[1])+start_chr, int(position_dict.get(feat)[2])+start_chr):
            if bp in dna_dict:
                if dna_dict[bp] == ['noncoding', None]:
                    dna_dict[bp] = [feat, feature_type]
            else: 
                dna_dict[bp]=[feat, feature_type]


    return(dna_dict)


def intergenic_regions(chrom,start_chr,dna_dict):
    """Getting intergenic regions from chromosome of interest 

    Parameters
    ----------
    chrom : str
        Name of the chromosome in roman where to extract the information.
    start_chr : int
       2nd output of the gene_location function 
    dna_dict : dict 
        1st output of the gene_location function 

    Returns
    -------
    dna_dict_new : dict
        
    genomicregions_list: list

    
    """


    sgd_features_file=load_sgd_tab()
    ## GET FEATURES FROM INTERGENIC REGIONS 

    genomicregions_list = sgd_features(sgd_features_file)[0]

    i = 2
    for genomicregion in genomicregions_list[1:]:
        dna_dict_new = feature_position(sgd_features(sgd_features_file)[i], chrom, start_chr, dna_dict, genomicregion)
        i += 1
    
    return dna_dict_new,genomicregions_list


def checking_features(feature_orf_dict,chrom,gene_position_dict,verbose):
    """ Checking input values 

    Parameters
    ----------
    feature_orf_dict : dict 
        last output of the gene_location function 
    chrom : str
        Name of the chromosome in roman where to extract the information.
    gene_position_dict : dict
        output of the read_pergene_file function 
    verbose : bool 
        If True it allows for warning messages 
    """

    ### TEST IF ELEMENTS IN FEATURE_ORF_DICT FOR SELECTED CHROMOSOME ARE THE SAME AS THE GENES IN GENE_POSITION_DICT BY CREATING THE DICTIONARY FEATURE_POSITION_DICT CONTAINING ALL THE GENES IN FEATURE_ORF_DICT WITH THEIR CORRESPONDING POSITION IN THE CHROMOSOME
    _,_,gene_information_file=load_default_files()

    gene_alias_dict = gene_aliases(gene_information_file)[0]
    orf_position_dict = {}
    for feature in feature_orf_dict:
        if feature_orf_dict.get(feature)[5] == chrom:
            if feature in gene_position_dict:
                orf_position_dict[feature] = [feature_orf_dict.get(feature)[6], feature_orf_dict.get(feature)[7]]
            else:
                for feature_alias in gene_alias_dict.get(feature):
                    if feature_alias in gene_position_dict:
                        orf_position_dict[feature_alias] = [feature_orf_dict.get(feature)[6], feature_orf_dict.get(feature)[7]]



    if sorted(orf_position_dict) == sorted(gene_position_dict):
        if verbose == True:
            print('Everything alright, just ignore me!')
        
    else:
        print('WARNING: Genes in feature_list are not the same as the genes in the gene_position_dict. Please check!')

    
def build_dataframe(dna_dict,start_chr,end_chr,insrt_in_chrom_list,reads_in_chrom_list,genomicregions_list,chrom):
    """Main function that build the big dataframe with all genes characteristics 

    Parameters
    ----------
    dna_dict : dict
        1st output of the function intergenic_regions
    start_chr : int
        2nd output of the function gene_location
    end_chr : int
        3rd output of the function gene_location
    insrt_in_chrom_list : list
        1st output of the function read_wig_file
    reads_in_chrom_list : list
        2nd output of the function read_wig_file
    genomicregions_list : list
        All the annotated genomic regions, 2nd output of the intergenic_regions function
    chrom : str
        Name of the chromosome in roman where to extract the information.

    Returns
    -------
    dataframe
    - dna_df2: Dataframe containing information about the selected chromosome. This includes the following columns:
        - Feature name
        - Standard name of the feature
        - Aliases of feature name (if any)
        - Feature type (e.g. gene, telomere, centromere, etc. If None, this region is not defined)
        - Chromosome
        - Position of feature type in terms of bp relative to chromosome.
        - Length of region in terms of basepairs
        - Number of insertions in region
        - Number of insertions in truncated region where truncated region is the region without the first and last 100bp.
        - Number of reads in region
        - Number of reads in truncated region.
        - Number of reads per insertion (defined by Nreads/Ninsertions)
        - Number of reads per insertion in truncated region (defined by Nreads_truncatedgene/Ninsertions_truncatedgene)
        NOTE: truncated regions are only determined for genes. For the other regions the truncated region values are the same as the non-truncated region values.


        
    """

    _,essentials_file,gene_information_file=load_default_files()
    

    ## DETERMINE THE NUMBER OF TRANSPOSONS PER BP FOR EACH FEATURE

    reads_loc_list = [0] * len(dna_dict) # CONTAINS ALL READS JUST LIKE READS_IN_CHROM_LIST, BUT THIS LIST HAS THE SAME LENGTH AS THE NUMBER OF BP IN THE CHROMOSOME WHERE THE LOCATIONS WITH NO READS ARE FILLED WITH ZEROS
    i = 0
    
    for ins in insrt_in_chrom_list:
        if len(reads_loc_list) > ins-1: # there are some insertions outside  the chromosome that gives an error 
            reads_loc_list[ins-1] = reads_in_chrom_list[i]
            i += 1
    

    feature_NameAndType_list = []
    f_previous = dna_dict.get(start_chr)[0]
    f_type = dna_dict.get(start_chr)[1]
    N_reads = []
    N_reads_list_true=[]
    N_reads_list = []
    N_reads_truncatedgene_list = []
    N_insrt_truncatedgene_list = []
    N_insrt_list = []
    N_bp = 1
    N_bp_list = []
    f_start = 0
    f_end = 0
    f_pos_list = []
    i = 0
    for bp in dna_dict:
        f_current = dna_dict.get(bp)[0]
        if f_current == f_previous:
            f_type = dna_dict.get(bp)[1]
            f_end += 1
            N_bp += 1
            N_reads.append(reads_loc_list[i])
        elif (f_current != f_previous or (i+start_chr) == end_chr):# and not f_current.endswith('-A'):
            feature_NameAndType_list.append([f_previous, f_type])
            N_reads_list.append(sum(N_reads))
            N_reads_list_true.append(np.array(N_reads,dtype=float))
            N_insrt_list.append(len([ins for ins in N_reads if not ins == 0]))
            if not f_type == None and f_type.startswith('Gene'):
                N10percent = 100#int(len(N_reads) * 0.1) #TRUNCATED GENE DEFINITION
                N_reads_truncatedgene_list.append(sum(N_reads[N10percent:-N10percent]))
                N_insrt_truncatedgene_list.append(len([ins for ins in N_reads[N10percent:-N10percent] if not ins == 0]))
            else:
                N_reads_truncatedgene_list.append(sum(N_reads))
                N_insrt_truncatedgene_list.append(len([ins for ins in N_reads if not ins == 0]))

            N_bp_list.append(N_bp)
            N_reads = []
            N_bp = 1
            f_pos_list.append([f_start, f_end+f_start])
            f_start = f_start + f_end + 1
            f_end = 0
            f_previous = f_current
        i += 1

    N_reads_per_ins_list = []
    N_reads_per_ins_truncatedgene_list = []
    for i in range(len(N_reads_list)):
        if N_insrt_list[i] < 5: # upper bound of low number of transposons 
            N_reads_per_ins_list.append(0)
            N_reads_per_ins_truncatedgene_list.append(0)
        elif N_insrt_truncatedgene_list[i] < 5:
            N_reads_per_ins_list.append(N_reads_list[i]/(N_insrt_list[i]-1))
            N_reads_per_ins_truncatedgene_list.append(0)
        else:
            N_reads_per_ins_list.append(N_reads_list[i]/(N_insrt_list[i]-1))
            N_reads_per_ins_truncatedgene_list.append(N_reads_truncatedgene_list[i]/N_insrt_truncatedgene_list[i])


    #############get all essential genes together with their aliases##############
    with open(essentials_file, 'r') as f:
        essentials_temp_list = f.readlines()[1:]
    essentials_list = [essential.strip('\n') for essential in essentials_temp_list]
    del essentials_temp_list

    gene_alias_dict = gene_aliases(gene_information_file)[0]
    for key, val in gene_alias_dict.items():
        if key in essentials_list:
            for alias in val:
                essentials_list.append(alias)

    #ADD
    essentiality_list = []
    for feature in feature_NameAndType_list:
        if not feature[0] == "noncoding":
            if feature[1] in genomicregions_list:
                essentiality_list.append(None)
            elif feature[0] in essentials_list:
                essentiality_list.append(True)
            else:
                essentiality_list.append(False)
        else:
            essentiality_list.append(None)

    

    feature_name_list = []
    feature_type_list = []
    feature_alias_list = []
    feature_standardname_list = []
    for feature_name in feature_NameAndType_list:
        feature_name_list.append(feature_name[0])
        feature_type_list.append(feature_name[1])
        if feature_name[1] != None and feature_name[1].startswith('Gene') and feature_name[0] in gene_alias_dict:
            if gene_alias_dict.get(feature_name[0])[0] == feature_name[0]:
                feature_standardname_list.append(feature_name[0])
                feature_alias_list.append('')
            else:
                if len(gene_alias_dict.get(feature_name[0])) > 1:
                    feature_standardname_list.append(gene_alias_dict.get(feature_name[0])[0])
                    feature_alias_list.append(gene_alias_dict.get(feature_name[0])[1:])
                else:
                    feature_standardname_list.append(gene_alias_dict.get(feature_name[0])[0])
                    feature_alias_list.append('')
        else:
            feature_standardname_list.append(feature_name[0])
            feature_alias_list.append('')


    all_features = {'Feature_name': feature_name_list,
                    'Standard_name': feature_standardname_list,
                    'Feature_alias':feature_alias_list,
                    'Feature_type': feature_type_list,
                    'Essentiality': essentiality_list,
                    'Chromosome': [chrom]*len(feature_name_list),
                    'Position': f_pos_list,
                    'Nbasepairs':N_bp_list,
                    'Ninsertions':N_insrt_list,
                    'Ninsertions_truncatedgene':N_insrt_truncatedgene_list,
                    'Nreads':N_reads_list,
                    'Nreads_list':  N_reads_list_true,
                    'Nreads_truncatedgene':N_reads_truncatedgene_list,
                    'Nreadsperinsrt':N_reads_per_ins_list,
                    'Nreadsperinsrt_truncatedgene':N_reads_per_ins_truncatedgene_list}


    dna_df2 = pd.DataFrame(all_features, columns = [column_name for column_name in all_features]) #search for feature using: dna_df2.loc[dna_df2['Feature'] == 'CDC42']
    #CREATE NEW COLUMN WITH ALL DOMAINS OF THE GENE (IF PRESENT) AND ANOTHER COLUMN THAT INCLUDES LISTS OF THE BP POSITIONS OF THESE DOMAINS
    return dna_df2



from ..utils import chromosomename_roman_to_arabic
import pkg_resources
import os

def chromosome_position(gff_file):
    """Get the start and end position of each chromosome and determine their respective length.
    Input is a .gff file downloaded from https://www.ensembl.org/Saccharomyces_cerevisiae/Info/Index
    Output are three dictionaries for length, start and end position. All dictionaries have keys representing the chromosome number in roman numerals.
    To get all dictionaries, use: 'a,b,c, chromosome_and_gene_position.chromosome_position()'.
    'a' = chromosome length
    'b' = chromosome start position
    'c' = chromosome end position

    Parameters
    ----------
    gff_file : str
        The file path of  a .gff file downloaded from https://www.ensembl.org/Saccharomyces_cerevisiae/Info/Index

    Returns
    -------
    dict 
        A dictionary relating each chromosome with its length
    dict 
        A dictionary relating each chromosome with its start position
    dict 
        A dictionary relating each chromosome with its end  position
    """    

    if gff_file==None:
        default_path = pkg_resources.resource_filename("transposonmapper", "data_files/")
        gff_file = os.path.join(
            default_path, "Saccharomyces_cerevisiae.R64-1-1.99.gff3"
        )
    # Get roman to arabic dictionary
    roman_to_arabic = chromosomename_roman_to_arabic()[1]

    # GET END POSITIONS OF THE CHROMOSOMES FROM THE GFF FILE AND STORE THEM IN A DICTIONARY
    chr_length_dict = {}
    chr_length_list = []
    with open(gff_file) as f:
        line_counter = 0
        next(f)
        while line_counter < 17:
            lines = f.readline()
            chr_line_list = lines.strip("\n").replace(" ", "\t").split("\t")
            chr_number = chr_line_list[3]
            # if chr_number != 'Mito':
            chr_length = int(chr_line_list[5])
            chr_length_list.append(chr_length)
            chr_length_dict[chr_number] = chr_length
            line_counter += 1

    # DETERMINE START AND END POSITION OF EACH OF THE CHROMOSOMES
    chr_start_pos_dict = {}
    chr_end_pos_dict = {}
    counter = 0
    for roman in roman_to_arabic.keys():
        chr_start_pos_dict[roman] = sum(chr_length_list[:counter]) + 1
        chr_end_pos_dict[roman] = sum(chr_length_list[: counter + 1])
        if roman == "I":
            chr_start_pos_dict[roman] = 1
        counter += 1

    return (chr_length_dict, chr_start_pos_dict, chr_end_pos_dict)



import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import numpy as np 

from transposonmapper.importing import load_default_files
from transposonmapper.processing import list_known_essentials

def profile_genome_plot(bar_width,l_genome,allinsertionsites_list,allcounts_binnedlist,summed_chr_length_dict,
                         middle_chr_position,chrom_list,variable,genes_currentchrom_pos_list,gene_pos_dict):
    """Plot function to show the whole insertion map throughout the genome 

    Parameters
    ----------
    bar_width : int
        The width for the histogram of the plot, by default None , which means internally the length of the genome over 1000 
    l_genome : int
        The length of the genome in bp
    allinsertionsites_list : list
        List of insertions sites 
    allcounts_binnedlist : list
        List of binned counts 
    summed_chr_length_dict : dict
        The cumulative sum of the length of every chromosome 
    middle_chr_position : dict
        Middle chromosome position per chromosome
    chrom_list : list
        A list of all the chromosomes
    variable : str
        It could be "transposons" or "reads"
    genes_currentchrom_pos_list : list
        List of genes per chromosome 
    gene_pos_dict : dict 
        Postion along the genome of every gene 
    """    

    _,essential_file,_=load_default_files()
    genes_essential_list=list_known_essentials(essential_file)


    plt.figure(figsize=(19.0,9.0))
    grid = plt.GridSpec(20, 1, wspace=0.0, hspace=0.0)

    textsize = 12
    textcolor = "#000000"
    barcolor= "#333333"
    chrom_color=(0.9,0.9,0.9,1.0)
    essential_face_color="#00F28E"
    non_essential_face_color="#F20064"
    alpha=0.8
    
    binsize = bar_width
    ax = plt.subplot(grid[0:19,0])
    ax.grid(False)
    ax.tick_params(axis='x', which='major', pad=30)
    
    
    
    axc = plt.subplot(grid[19,0])
    
    axc.set_xlim(0,l_genome)
    axc.tick_params(
        axis='x',          # changes apply to the x-axis
        which='both',      # both major and minor ticks are affected
        bottom=False,      # ticks along the bottom edge are off
        top=False,         # ticks along the top edge are off
        labelbottom=False) # labels along the bottom edge are off

    axc.tick_params(
        axis='y',          # changes apply to the y-axis
        which='both',      # both major and minor ticks are affected
        left=False,        # ticks along the bottom edge are off
        right=False,       # ticks along the top edge are off
        labelleft=False)   # labels along the bottom edge are off
    
    ax.set_xlim(0,l_genome)
    
    # bar lines
    ax.bar(allinsertionsites_list,allcounts_binnedlist,width=binsize,color=barcolor)
   
    
    # chromosome lines
    for chrom in summed_chr_length_dict:
        ax.axvline(x = summed_chr_length_dict.get(chrom), linestyle='-', color=chrom_color)

    ax.set_xticks(middle_chr_position)
    ax.set_xticklabels(chrom_list, fontsize=textsize)
    
    # Axis labels
    if variable == "transposons":
        ax.set_ylabel('Transposon Count', fontsize=textsize, color=textcolor)
    elif variable == "reads":
        ax.set_ylabel('Read Count', fontsize=textsize, color=textcolor)

    # colored bars in the bottom 
    for gene in genes_currentchrom_pos_list:
        if not gene_pos_dict.get(gene)[0] == 'Mito':
            gene_start_pos = summed_chr_length_dict.get(gene_pos_dict.get(gene)[0]) + int(gene_pos_dict.get(gene)[1])
            gene_end_pos = summed_chr_length_dict.get(gene_pos_dict.get(gene)[0]) + int(gene_pos_dict.get(gene)[2])
            if gene in genes_essential_list:
                axc.axvspan(gene_start_pos,gene_end_pos,facecolor=essential_face_color,alpha=alpha)
            else:
                axc.axvspan(gene_start_pos,gene_end_pos,facecolor=non_essential_face_color,alpha=alpha)
    



def save_per_gene(filename, tn_per_gene, reads_per_gene, aliases_designation):
    """Create text file with transposons and reads per gene
 
    NOTE THAT THE TRANSPOSON WITH THE HIGHEST READ COUNT IS IGNORED.
    E.G. IF THIS FILE IS COMPARED WITH THE _PERGENE_INSERTIONS.TXT FILE THE 
    READS DON'T ADD UP (SEE https://groups.google.com/forum/#!category-topic/satayusers/bioinformatics/uaTpKsmgU6Q)
    TOO REMOVE THIS HACK, CHANGE THE INITIALIZATION OF THE VARIABLE readpergene

    Parameters
    ----------
    filename : str
        Path with the filename extension included(e.g "data_file/file.txt") 
        describing where do you want to store the results.
        By default it will be stored in the same location as the bamfile, 
        with the same basename. 
        Example, if the bamfile path is data_file/data_1.bam then the  file will be data_file/data_1.bam_pergene.txt
    tn_per_gene : int 
        The number of transposons found per gene 
    reads_per_gene : int
        The number of reads found per gene 
    aliases_designation : dict
        Last output of the function read_genes  _, _, aliases_designation = read_genes(
        gff_file, essential_file, gene_name_file)
    """

    with open(filename, "w") as f:

        f.write("Gene name\tNumber of transposons per gene\tNumber of reads per gene\n")

        for gene in tn_per_gene:
            tnpergene = tn_per_gene[gene]
            readpergene = reads_per_gene[gene]
            if gene in aliases_designation:
                gene_alias = aliases_designation.get(gene)[0]
            else:
                gene_alias = gene
            f.write(gene_alias + "\t" + str(tnpergene) + "\t" + str(readpergene) + "\n")




from .get_coordinates_genes import get_coordinates_genes
from .get_chromosome_properties import (
    get_chromosome_names,
    get_sequence_length,
    get_chromosome_reads,
)
from .get_chromosome_position import chromosome_position
from .get_gene_position import gene_position
from .list_gene_names import (list_gene_names)
from .gene_aliases import gene_aliases

__all__ = [
    "get_coordinates_genes",
    "get_chromosome_names",
    "get_sequence_length",
    "get_chromosome_reads",
    "list_gene_names",
    "chromosome_position",
    "gene_position",
    "gene_aliases"
]




# -*- coding: utf-8 -*-

from transposonmapper.importing import load_sgd_tab
from transposonmapper.utils   import  chromosomename_roman_to_arabic

def sgd_features(filepath=None):
    """This function read the file SGD_features.tab and create a dictionary with useful info for processing

    Parameters
    ----------
    filepath : str, optional
        filepath of the sgd.tab file , by default None

    Returns
    -------
    dict
         A dictionary with the following info: 
        key:
            0. Feature name
        value:
            0. feature type (l[1])
            1. feature qualifier (Verified or Dubious) (l[2])
            2. Standard name (l[4])
            3. Aliases (separated by '|') (l[5])
            4. Parent feature name (typically 'chromosome ...') (l[6])
            5. Chromosome (l[8])
            6. start coordinate (starting at 0 for each chromosome) (l[9])
            7. end coordinate (starting at 0 for each chromosome) (l[10])
           
            This file reads the SGD_features.txt file found at http://sgd-archive.yeastgenome.org/curation/chromosomal_feature/
    """

    if filepath == None:
        filepath=load_sgd_tab()

    arabic_to_roman_dict=chromosomename_roman_to_arabic()[0]
    
    with open(filepath) as f:
        lines = f.readlines()


    feature_list = []
    feature_orf_dict = {}
    feature_ars_dict = {}
    feature_telomere_dict = {}
    feature_ltr_dict = {}
    feature_centromere_dict = {}
    feature_Xelement_dict = {}
    feature_intron_dict = {}
    feature_ncrna_dict = {}
    feature_ncexon_dict = {}
    feature_trna_dict = {}
    feature_snorna_dict = {}
    feature_teg_dict = {}
    feature_5p_utrintron_dict = {}
    feature_mas_dict = {}
    feature_snrna_dict = {}
    feature_rrna_dict = {}
    feature_ets_dict = {}
    feature_its_dict = {}
    feature_oor_dict = {}
    feature_telrna_dict = {}
    
    for line in lines:
        l = line.strip('\n').split('\t')
        if not l[1] in feature_list:
            feature_list.append(l[1])

        if not l[8].endswith('micron') and not l[8] == '':
            chromosome = arabic_to_roman_dict.get(int(l[8]))
            if l[1] == 'ORF':
                feature_orf_dict[l[3]] = [l[1], l[2], l[4], l[5], l[6], chromosome, l[9],l[10]]
            elif l[1] == 'ARS':
                feature_ars_dict[l[3]] = [l[1], l[2], l[4], l[5], l[6], chromosome, l[9],l[10]]
            elif l[1] == 'telomere':
                feature_telomere_dict[l[3]] = [l[1], l[2], l[4], l[5], l[6], chromosome, l[9],l[10]]
            elif l[1] == 'long_terminal_repeat':
                feature_ltr_dict[l[3]] = [l[1], l[2], l[4], l[5], l[6], chromosome, l[9],l[10]]
            elif l[1] == 'centromere':
                feature_centromere_dict[l[3]] = [l[1], l[2], l[4], l[5], l[6], chromosome, l[9],l[10]]
            elif l[1] == 'X_element':
                feature_Xelement_dict[l[3]] = [l[1], l[2], l[4], l[5], l[6], chromosome, l[9],l[10]]
            elif l[1] == 'intron':
                feature_intron_dict[l[6]] = [l[1], l[2], l[4], l[5], l[6], chromosome, l[9],l[10]]
            elif l[1] == 'ncRNA_gene':
                feature_ncrna_dict[l[3]] = [l[1], l[2], l[4], l[5], l[6], chromosome, l[9],l[10]]
            elif l[1] == 'noncoding_exon':
                feature_ncexon_dict[l[6]] = [l[1], l[2], l[4], l[5], l[6], chromosome, l[9],l[10]]
            elif l[1] == 'tRNA_gene':
                feature_trna_dict[l[3]] = [l[1], l[2], l[4], l[5], l[6], chromosome, l[9],l[10]]
            elif l[1] == 'snoRNA_gene':
                feature_snorna_dict[l[3]] = [l[1], l[2], l[4], l[5], l[6], chromosome, l[9],l[10]]
            elif l[1] == 'transposable_element_gene':
                feature_teg_dict[l[3]] = [l[1], l[2], l[4], l[5], l[6], chromosome, l[9],l[10]]
            elif l[1] == 'five_prime_UTR_intron':
                feature_5p_utrintron_dict[l[6]] = [l[1], l[2], l[4], l[5], l[6], chromosome, l[9],l[10]]
            elif l[1] == 'matrix_attachment_site':
                feature_mas_dict[l[3]] = [l[1], l[2], l[4], l[5], l[6], chromosome, l[9],l[10]]
            elif l[1] == 'snRNA_gene':
                feature_snrna_dict[l[3]] = [l[1], l[2], l[4], l[5], l[6], chromosome, l[9],l[10]]
            elif l[1] == 'rRNA_gene':
                feature_rrna_dict[l[3]] = [l[1], l[2], l[4], l[5], l[6], chromosome, l[9],l[10]]
            elif l[1] == 'external_transcribed_spacer_region':
                feature_ets_dict[l[6]] = [l[1], l[2], l[4], l[5], l[6], chromosome, l[9],l[10]]
            elif l[1] == 'internal_transcribed_spacer_region':
                feature_its_dict[l[6]] = [l[1], l[2], l[4], l[5], l[6], chromosome, l[9],l[10]]
            elif l[1] == 'origin_of_replication':
                feature_oor_dict[l[3]] = [l[1], l[2], l[4], l[5], l[6], chromosome, l[9],l[10]]
            elif l[1] == 'telomerase_RNA_gene':
                feature_telrna_dict[l[3]] = [l[1], l[2], l[4], l[5], l[6], chromosome, l[9],l[10]]


    


    genomicregions_list = ['ORF', 'ARS', 'Telomere', 'long_terminal_repeat',
                           'Centromere', 'X_element', 'Intron', 'ncRNA_gene',
                           'Noncoding_exon', 'tRNA_gene', 'snoRNA_gene',
                           'transposable_element_gene', 'five_prime_UTR_intron',
                           'matrix_attachment_site', 'snRNA_gene', 'rRNA_gene',
                           'external_transcribed_spacer_region',
                           'internal_transcribed_spacer_region',
                           'origin_of_replication', 'telomerase_RNA_gene']


    return(genomicregions_list, feature_orf_dict, feature_ars_dict, feature_telomere_dict,
           feature_ltr_dict, feature_centromere_dict, feature_Xelement_dict, feature_intron_dict,
           feature_ncrna_dict, feature_ncexon_dict, feature_trna_dict,
           feature_snorna_dict, feature_teg_dict, feature_5p_utrintron_dict,
           feature_mas_dict, feature_snrna_dict, feature_rrna_dict,
           feature_ets_dict, feature_its_dict, feature_oor_dict,
           feature_telrna_dict)





        
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import numpy as np 
        
def dna_features_plot(chrom, dna_df2,roi_start,roi_end,len_chr,plotting,variable):  
    """Plot function for the genomic features per chromosome 

    Parameters
    ----------
    chrom : str
        - Region: e.g. chromosome number (either a normal number between 1 and 16 or in roman numerals between I and XVI),
         a list like ['V', 0, 14790] which creates a barplot between basepair 0 and 14790) or a genename.
    dna_df2 : pandas.dataframe 
        Dataframe containing information about the selected chromosome
    roi_start : int
        The start of the chromosome
        roi_start,roi_end,region_type,chrom=input_region(region=region,verbose=verbose)
    roi_end : int
        The end of the chromosome
        roi_start,roi_end,region_type,chrom=input_region(region=region,verbose=verbose)
    len_chr : [type]
        Length of the  chromosome
        dna_dict,start_chr,end_chr,len_chr,feature_orf_dict=gene_location(chrom,gene_position_dict,verbose)
    plotting : optional
        default True
    variable : optional 
        default "reads"
    
    """ 


    if plotting == True:
        create_plottitle = chrom
        noncoding_color = "#002538"
        essential_color = "#10e372"
        nonessential_color = "#d9252e"
        codingdna_color = '#29a7e6'
        textcolor = "#000000"

        textsize = 14

        plt.figure(figsize=(19,9))
        grid = plt.GridSpec(20, 1, wspace=0.0, hspace=0.01)

        axc = plt.subplot(grid[19,0])

        axc.tick_params(labelsize=textsize)
        axc.set_yticklabels([])
        axc.tick_params(
            axis='x',          # changes apply to the x-axis
            which='both',      # both major and minor ticks are affected
            bottom=False,      # ticks along the bottom edge are off
            top=False,         # ticks along the top edge are off
            labelbottom=False) # labels along the bottom edge are off

        axc.tick_params(
            axis='y',          # changes apply to the y-axis
            which='both',      # both major and minor ticks are affected
            left=False,        # ticks along the bottom edge are off
            right=False,       # ticks along the top edge are off
            labelleft=False)   # labels along the bottom edge are off


        ax = plt.subplot(grid[0:19,0])
        

        ax.grid(linestyle='-', alpha=1.0)
        ax.tick_params(labelsize=textsize)

        ax.tick_params(axis='x', which='major', pad=30)
        ax.ticklabel_format(axis='x', style='sci', scilimits=(0,0))
        ax.xaxis.get_offset_text().set_fontsize(textsize)
        ax.set_xlabel("Basepair position on chromosome "+chrom, fontsize=textsize, color=textcolor, labelpad=10)
        ax.set_title(create_plottitle, fontsize=textsize, color=textcolor)


        feature_middle_pos_list = []
        sum_bp = 0
        for x in dna_df2['Nbasepairs']:
            feature_middle_pos_list.append(x/2 + sum_bp)
            sum_bp += x
        
        feature_width_list = list(dna_df2['Nbasepairs'])


        barcolor_list = []
        for feature in dna_df2['Feature_name']:
            if feature == 'noncoding':
                barcolor_list.append(noncoding_color)
            elif dna_df2.loc[dna_df2['Feature_name'] == feature]['Essentiality'].iloc[0] == False:
                barcolor_list.append(nonessential_color)
            elif dna_df2.loc[dna_df2['Feature_name'] == feature]['Essentiality'].iloc[0] == True:
                barcolor_list.append(essential_color)
            elif dna_df2.loc[dna_df2['Feature_name'] == feature]['Essentiality'].iloc[0] == None:
                barcolor_list.append(codingdna_color)

        

        legend_noncoding = mpatches.Patch(color=noncoding_color, label="Noncoding DNA")
        legend_essential = mpatches.Patch(color=essential_color, label="Annotated essential genes")
        legend_nonessential = mpatches.Patch(color=nonessential_color, label="Nonessential genes")
        legend_coding = mpatches.Patch(color=codingdna_color, label="Other genomic regions")
        leg = ax.legend(handles=[legend_noncoding, legend_essential, legend_nonessential, legend_coding]) #ADD

        for text in leg.get_texts():
            text.set_color(textcolor)
        
        if variable == "insertions":
            ax.bar(feature_middle_pos_list, list(dna_df2['Ninsertions']), feature_width_list, color=barcolor_list)
            ax.set_ylabel("Transposons per region", fontsize=textsize, color=textcolor)
        elif variable == "reads":
            ax.bar(feature_middle_pos_list, list(dna_df2['Nreads']), feature_width_list, color=barcolor_list)
            ax.set_ylabel("Reads per region", fontsize=textsize, color=textcolor)


        if roi_start != None and roi_end != None and roi_start < len_chr and roi_end < len_chr:
            ax.set_xlim(roi_start, roi_end)
        else:
            ax.set_xlim(0, len_chr)



        

        l = 0
        counter = 0
        for width in feature_width_list:
            if dna_df2.loc[counter][4] == True:
                axc.axvspan(l,l+width,facecolor=essential_color,alpha=0.3)
            elif dna_df2.loc[counter][4] == False and not dna_df2.loc[counter][0] == 'noncoding':
                axc.axvspan(l,l+width,facecolor=nonessential_color,alpha=0.3)
            elif dna_df2.loc[counter][4] == None and not dna_df2.loc[counter][0] == 'noncoding':
                axc.axvspan(l,l+width,facecolor=codingdna_color,alpha=0.5)
            l += width
            counter += 1
        if roi_start != None and roi_end != None and roi_start < len_chr and roi_end < len_chr:
            axc.set_xlim(roi_start, roi_end)
        else:
            axc.set_xlim(0, len_chr)
        
        


import warnings


def get_chromosome_names(bam):
    """ This functions translate the format of the chromosome names
    from the alignment file from the pysam module into numbers
    

    Parameters
    ----------
    bam : dict , It is the output of the function  pysam.AlignmentFile(bamfile, "rb")    

    Returns
    -------
    ref_tid : dict, where the values are the chromosome numbers per key in the bam file. 
    
    """

    # ref_tid = {str(name): int(bam.get_tid(name)) + 1 for name in bam.get_reference_name}

    ref_tid = {} # 'I' | 0, 'II' | 1, ...
    for i in range(bam.nreferences): #if bam.nreferences does not work, use range(17) #16 chromosomes and the mitochondrial chromosome
        ref_name = bam.get_reference_name(i)
        ref_tid[ref_name] = bam.get_tid(ref_name)

    return ref_tid


def get_sequence_length(bam):
    """ This function returns the length of each chromosome and a cumulative sum of them.
    
    Parameters
    ----------
    bam : dict, It is the output of the function  pysam.AlignmentFile(bamfile, "rb")          

    Returns
    -------
    chr_lengths : dict, A dictionary which very value is the length in basepairs
    of each chromosome in the alignment file. 
    chr_lengths_cumsum : dict , A dictionary which very value is the sum of the length in basepairs
    of the previous chromosomes in the alignment file. chr_lengths_cumsum{n}=cumulative sum (n-1 previous chromosomes
    length)

    """

    chr_lengths = {}  # 'I' | 230218, 'II' | 813184, ...
    chr_lengths_cumsum = {}  # 'I' | 0, 'II' | 230218, 'III' |  1043402, ...
    ref_summedlength = 0
    ref_tid = get_chromosome_names(bam)
    for key in ref_tid.keys():
        ref_length = bam.get_reference_length(key)
        chr_lengths[key] = bam.get_reference_length(key)
        chr_lengths_cumsum[key] = ref_summedlength
        ref_summedlength += ref_length

    return chr_lengths, chr_lengths_cumsum


def get_chromosome_reads(bam):
    """ This function returns statistics about mapped/unmapped reads per chromosome as they are stored in the index.
    It makes use of the method get_index_statistics() from the pysam module. 
    Parameters
    ----------
    bam : dict, It is the output of the function  pysam.AlignmentFile(bamfile, "rb")              

    Returns
    -------
    mapped_reads : dict[str, list]
        Syntax 'I' | [mapped, unmapped, total reads]

    """
    stats = bam.get_index_statistics()
    mapped_reads = {}
    for stat in stats:
        mapped_reads[stat[0]] = [stat[1], stat[2], stat[3]]
        if stat[2] != 0:
            warnings.warn("Unmapped reads found in chromosome " + stat[0])

    return mapped_reads



import numpy as np
import timeit

from transposonmapper.mapping.find_chromosome_reads import find_chromosome_reads
from transposonmapper.mapping.correct_read_position import correct_read_position

from transposonmapper.properties import (
    get_chromosome_names,
    get_sequence_length,
    get_chromosome_reads,
)


def get_reads(bam):

    """This function retrieves  all reads within a specified genomic region. 
     
    
    Parameters
    ----------
    bam :  The output for the function pysam.AlignmentFile(bamfile, "rb")


    Returns
    -------
    numpy.array
        reads per genomic region
    numpy.array
        Array of three columns where the 2nd one indicated the start position where there was a transposon
    numpy.array
         A copy from the 2nd output 

    """

    # Get chromosome properties
    ref_tid = get_chromosome_names(bam)
    ref_names = list(ref_tid.keys())
    chr_lengths, _ = get_sequence_length(bam)
    chr_mapped_reads = get_chromosome_reads(bam)

    ### GET ALL READS WITHIN A SPECIFIED GENOMIC REGION
    tnnumber_dict = {}
    unique_insertions = 0

    for chr_num in ref_names:
        timer_start = timeit.default_timer()

        print("Getting reads for chromosome %s ..." % chr_num)
        chromosome = bam.fetch(chr_num, 0, chr_lengths[chr_num], until_eof=True)
        N_reads = chr_mapped_reads[chr_num][2]

        [flags, start_position, readlength] = find_chromosome_reads(chromosome, N_reads)

        start_position_corrected, flags_corrected = correct_read_position(
            flags, start_position, readlength
        )

        # CREATE ARRAY OF START POSITION AND FLAGS OF ALL READS IN GENOME
        ref_tid_kk = int(ref_tid[chr_num] + 1)
        if unique_insertions == 0:
            tncoordinates_array = np.array([])

        unique_insertions_per_read = 0  # Number of unique reads per insertion
        num_transposons = 1  # Number of unique reads in current chromosome (Number of transposons in current chromosome)
        for ii in range(1, len(start_position_corrected)):
            if (
                abs(start_position_corrected[ii] - start_position_corrected[ii - 1])
                <= 2
                and flags_corrected[ii] == flags_corrected[ii - 1]
            ):  # If two subsequent reads are within two basepairs and have the same orientation, add them together.
                unique_insertions_per_read += 1
            else:
                avg_start_pos = abs(
                    round(
                        np.mean(
                            start_position_corrected[
                                ii - unique_insertions_per_read - 1 : ii
                            ]
                        )
                    )
                )
                if tncoordinates_array.size == 0:  # include first read
                    tncoordinates_array = np.array(
                        [ref_tid_kk, int(avg_start_pos), int(flags_corrected[ii - 1])]
                    )
                    readnumb_list = [unique_insertions_per_read + 1]
                else:
                    tncoordinates_array = np.vstack(
                        (
                            tncoordinates_array,
                            [
                                ref_tid_kk,
                                int(avg_start_pos),
                                int(flags_corrected[ii - 1]),
                            ],
                        )
                    )
                    readnumb_list.append(unique_insertions_per_read + 1)
                unique_insertions_per_read = 0
                num_transposons += 1
                unique_insertions += 1

            if ii == len(start_position_corrected) - 1:  # include last read
                avg_start_pos = abs(
                    round(
                        np.mean(
                            start_position_corrected[
                                ii - unique_insertions_per_read - 1 : ii
                            ]
                        )
                    )
                )
                tncoordinates_array = np.vstack(
                    (
                        tncoordinates_array,
                        [ref_tid_kk, int(avg_start_pos), int(flags_corrected[ii - 1])],
                    )
                )
                readnumb_list.append(unique_insertions_per_read + 1)

        tnnumber_dict[chromosome] = num_transposons

        timer_end = timeit.default_timer()
        print(
            "Chromosome %s completed in %.3f seconds"
            % (chromosome, (timer_end - timer_start))
        )
        print("")

    readnumb_array = np.array(readnumb_list)

    tncoordinatescopy_array = np.array(tncoordinates_array, copy=True)

    return readnumb_array, tncoordinates_array, tncoordinatescopy_array




import pkg_resources
import os
def gene_aliases(gene_information_file=None):
    """Create three dictionaries containing aliases for genes
    Input is the path to 'Protein_Names.txt' file downloaded from https://www.uniprot.org/docs/yeast.
    If no input is given the file is automatically searched for at thisscriptlocation/../Data_Files/Yeast_Protein_Names.txt.
    Output is three dictionaries:
    aliases_designation_dict = gene aliases for common names (e.g. Bem1 and Sro1)
    aliases_sgd_dict = gene aliases for the search names in SGD (e.g. Bem1 and S000000404)
    aliases_swissprot_dict = gene aliases for Swiss Prot.
    The keys of the dictionaries are the systematic names of the genes (e.g. YBR200W for Bem1)
    
    Search through lists to get corresponding key:
        [key for key, val in aliases.items() if 'TFC3' in val]

    Parameters
    ----------
    gene_information_file : str, optional
        Input is the path to 'Protein_Names.txt' file downloaded from https://www.uniprot.org/docs/yeast.
        If no input is given the file is automatically searched for at Yeast_Protein_Names.txt inside the package, by default None

    Returns
    -------
    dict 
        gene aliases for common names (e.g. Bem1 and Sro1)
    dict
        gene aliases for the search names in SGD (e.g. Bem1 and S000000404)
    dict 
         gene aliases for Swiss Prot

    """  

    if gene_information_file == None:

        default_path = pkg_resources.resource_filename("transposonmapper", "data_files/")
    
        gene_information_file = os.path.join(default_path, "Yeast_Protein_Names.txt")
        

        

    aliases_designation_dict = {}
    aliases_sgd_dict = {}
    aliases_swissprot_dict = {}

    with open(gene_information_file) as f:
        lines = f.readlines()

        for i in range(
            58, len(lines) - 6
        ):  # THE GENES START AT LINE 58 AND STOP 6 LINES BEFORE THE END OF THE FILE.
            l = lines[i]
            l_list = l.split()

            alias_counter = 0
            gene_designation_list = []
            for names in l_list:
                if names.endswith(";"):
                    gene_designation_list.append(names.strip(";"))
                    alias_counter += 1
            gene_designation_list.append(l_list[alias_counter])

            oln_name = l_list[alias_counter + 1]
            sgd_name = l_list[alias_counter + 4]
            swissprot_name = l_list[alias_counter + 2]
            if oln_name == "GAG" or oln_name == "POL":
                oln_name = l_list[alias_counter + 2]
                sgd_name = l_list[alias_counter + 5]
                swissprot_name = l_list[alias_counter + 3]

            aliases_designation_dict[oln_name] = gene_designation_list
            aliases_sgd_dict[oln_name] = sgd_name
            aliases_swissprot_dict[oln_name] = swissprot_name

    return (aliases_designation_dict, aliases_sgd_dict, aliases_swissprot_dict)




import numpy as np

def correct_read_position(flags, start_position, readlength):
    """Correct starting position for reads with reversed orientation

    Parameters
    ----------
    flags : numpy.array
        [description]
    start_position : numpy.array
        [description]
    readlength : numpy.array
        [description]

    Returns
    -------
    numpy.array 
        start position and flags corrected 
    """

    flag0coor_array = np.where(flags == 1)  # coordinates reads 5' -> 3'
    flag16coor_array = np.where(flags == -1)  # coordinates reads 3' -> 5'

    startdirect_array = start_position[flag0coor_array]
    flagdirect_array = flags[flag0coor_array]

    startindirect_array = (
        start_position[flag16coor_array] + readlength[flag16coor_array]
    )
    flagindirect_array = flags[flag16coor_array]

    start_position_corrected = np.concatenate(
        (startdirect_array, startindirect_array), axis=0
    )
    flags_corrected = np.concatenate((flagdirect_array, flagindirect_array), axis=0)

    start_sortindices = start_position_corrected.argsort(
        kind="mergesort"
    )  # use mergesort for stable sorting
    start_position_corrected = start_position_corrected[start_sortindices]
    flags_corrected = flags_corrected[start_sortindices]

    return start_position_corrected, flags_corrected



import numpy as np

from .samflag import samflags

def find_chromosome_reads(chromosome, N_reads: int):
    """Find the reads inside each chromosome 

    Parameters
    ----------
    chromosome :  str 
        The name of the chromosome , either in arabic numbers 1, 2 , or roman  "I", "II"
    N_reads : int
        

    Returns
    -------
    flags, start_position, readlength
        
    """
    # Initialize arrays
    start_position = np.empty(shape=(N_reads), dtype=int)
    flags = np.empty(shape=(N_reads), dtype=int)
    readlength = np.empty(shape=(N_reads), dtype=int)

    for index, reads in enumerate(chromosome):
        read = str(reads).split("\t")

        start_position[index] = int(read[3]) + 1

        # GET FLAG FOR EACH READ. IF READ ON FORWARD STRAND, ASSIGN VALUE 1, IF READ ON REVERSE STRAND ASSIGN VALUE -1, IF READ UNMAPPED OR SECONDARY ALIGNMENT ASSIGN VALUE 0
        #            flag_array[read_counter] = int(read[1])
        samprop = samflags(flag=int(read[1]), verbose=False)[1]
        if "read reverse strand" in samprop:
            flags[index] = -1
        else:
            flags[index] = 1
        if "not primary alignment" in samprop or "read unmapped" in samprop:
            flags[index] = 0

        cigarmatch_list = []
        if not reads.cigartuples == None:
            for cigar_type, cigar_length in reads.cigartuples:
                if cigar_type == 0:
                    cigarmatch_list.append(cigar_length)
                elif cigar_type == 2:
                    cigarmatch_list.append(cigar_length)
        match_length = sum(cigarmatch_list)

        readlength[index] = match_length  # int(len(read[9]))

    return flags, start_position, readlength




import numpy as np
import pandas as pd



from transposonmapper.statistics.dataframe_from_pergene_helpers import (essential_genes, 
                                                                        read_pergene_file, reads_per_insertion)

def dataframe_from_pergenefile(pergenefile, verbose=True):
    """This function creates a dataframe with the information from a pergene.txt file.
        
    
    The gene_essentiality is created based on the genes present in the Cerevisiae_EssentialGenes_List_1.txt and Cerevisiae_EssentialGenes_List_2.txt files
    The number of reads per insertion (Nreadsperinsrt) is determined by dividing the read_per_gene column by the tn_per_gene column.
    
    Author: Gregory van Beek

    Parameters
    ----------
    pergenefile : str
        absolute path to the pergene.txt file , one of the outputs of the transposonmapper module
    verbose : bool, optional
        [description], by default True

    Returns
    -----------
    dataframe
        Output is a dataframe where each row is a single gene and with the following columns:
            - gene_names
            - gene_essentiality
            - tn_per_gene
            - read_per_gene
            - Nreadsperinsrt
    """

   
# read file
    genenames_list,tnpergene_list,readpergene_list,lines=read_pergene_file(pergenefile)

    

# determine number of reads per insertion per gene
    readperinspergene_list=reads_per_insertion(tnpergene_list,readpergene_list,lines)

   

# determine essential genes
    geneessentiality_list=essential_genes(genenames_list,lines)


# create dataframe
    read_gene_dict = {"gene_names": genenames_list,
                      "gene_essentiality": geneessentiality_list,
                      "tn_per_gene": tnpergene_list,
                      "read_per_gene": readpergene_list,
                      "Nreadsperinsrt": readperinspergene_list}

    read_gene_df = pd.DataFrame(read_gene_dict, columns = [column_name for column_name in read_gene_dict])

    return(read_gene_df)





import os
import numpy as np


def save_as_wig(wigfile, tncoordinates_array, ref_tid, readnumb_array):
    """This function writes in a .wig file located in the same folder as the bamfile, the information abot how many transposons
    were found in each genomic location. 

    Parameters
    ----------
    wigfile : str 
        Path with the filename extension included(e.g "data_file/wigfile_name.wig") describing where do you want to store the results.
        By default it will be stored in the same location as the bamfile, with the same basename. 
        Example, if the bamfile path is data_file/data_1.bam then the wig file will be data_file/data_1.bam.wig
    tncoordinates_array : numpy array 
        Second Output from the get_reads.py function : _,tncoordinates_array,_=get_reads(bam)
    ref_tid : dict 
        Output from the get_chromosome_names function : ref_tid = get_chromosome_names(bam)
    readnumb_array : numpy array 
        1st Output from the get_reads.py function : readnumb_array,_,_=get_reads(bam)
    """

    readnumbwig_array = readnumb_array
    ref_names = list(ref_tid.keys())

    unique_index_array = np.array([], dtype=int)  # =cc
    N_uniques_perchr_list = []
    ll = 0
    for kk in ref_names:
        # get indices for current chromosome.
        index = np.where(tncoordinates_array[:, 0] == int(ref_tid[kk] + 1))

        # get all insertion locations (in tncoordinates, all rows, column 1)
        unique_index = np.unique(tncoordinates_array[index][:, 1], return_index=True)[1]

        unique_index_array = np.append(unique_index_array, (unique_index + ll), axis=0)

        # total amount unique indices found untill current chromosome
        ll += np.count_nonzero(tncoordinates_array[:, 0] == int(ref_tid[kk] + 1))
        N_uniques_perchr_list.append(ll)

    # Collect duplicates
    duplicate_list = []  # =dd
    ll = 0
    index_last_unique_previous_chromosome = 0
    for ii in N_uniques_perchr_list:
        index_last_unique = np.where(unique_index_array <= ii)[0][-1]
        for jj in range(ll, ii):
            if (
                int(jj)
                not in unique_index_array[
                    index_last_unique_previous_chromosome:index_last_unique
                ]
            ):
                duplicate_list.append(jj)
        index_last_unique_previous_chromosome = index_last_unique
        ll = ii

    # SUM READNUMB VALUES AT INDEX IN DUPLICATE_LIST AND DUPLICATE_LIST-1
    for ii in duplicate_list:
        readnumbwig_array[ii - 1] = readnumbwig_array[ii - 1] + readnumbwig_array[ii]

    tncoordinateswig_duplicatesremoved_array = np.delete(
        tncoordinates_array, duplicate_list, axis=0
    )
    readnumbwig_duplicatesremoved_array = np.delete(
        readnumbwig_array, duplicate_list, axis=0
    )

    # Write wigfile
    with open(wigfile, "w") as f:
        base = os.path.basename(wigfile)
        filename = os.path.splitext(base)[0]
        f.write("track type=wiggle_0 ,maxheightPixels=60 name=" + filename + "\n")
        for kk in ref_names:
            f.write("VariableStep chrom=chr" + kk + "\n")

            index = np.where(
                tncoordinateswig_duplicatesremoved_array[:, 0] == int(ref_tid[kk] + 1)
            )  # get indices for current chromosome.
            for ii in index[0]:
                f.write(
                    str(tncoordinateswig_duplicatesremoved_array[ii][1])
                    + " "
                    + str(readnumbwig_duplicatesremoved_array[ii])
                    + "\n"
                )



from .save_as_bed import save_as_bed
from .save_per_gene import save_per_gene
from .save_per_gene_insertions import (
    save_per_essential_insertions,
    save_per_gene_insertions,
)
from .save_as_wig import save_as_wig

__all__ = [
    "save_as_wig",
    "save_as_bed",
    "save_per_gene",
    "save_per_essential_insertions",
    "save_per_gene_insertions",
]




import os
import pkg_resources

def list_gene_names(gene_information_file=None):
    """Create a list of all known gene names and their aliases as listed on SGD (or as provided as an optional input file)
    Input is a standard file downloaded from https://www.uniprot.org/docs/yeast.
    Output is list of all genes, which also includes all the aliases (if they exists).

    Parameters
    ----------
    gene_information_file : str, optional
        Input is a standard file downloaded from https://www.uniprot.org/docs/yeast, by default None

    Returns
    -------
    list
        Output is list of all genes, which also includes all the aliases (if they exists).
    """   

    if gene_information_file == None:
       
        default_path = pkg_resources.resource_filename("transposonmapper", "data_files/")
    
        gene_information_file = os.path.join(
            default_path, "Yeast_Protein_Names.txt"
        )
        
    gene_name_list = []  # INLCUDES ALL GENE NAMES AND POTENTIAL ALIASES
    gene_oln_list = []  # INCLUDE ONLY THE OLN NAMING CONVENTION
    gene_sgd_list = []  # INCLUDE THE FIRST DESIGNATION NAME THAT IS IN THE INPUT LIST
    gene_counter = 0
    with open(gene_information_file) as f:
        lines = f.readlines()
        for i in range(
            58, len(lines) - 6
        ):  # THE GENES START AT LINE 58 AND STOP 6 LINES BEFORE THE END OF THE FILE.
            n = 0
            l = lines[i]

            extra_columns = l.count(
                ";"
            )  # COUNT HOW MANY TIMES ';' OCCURS IN A LINE. THIS IS NEEDED TO GET THE RIGHT COLUMNS AS SOMETIMES ALIASES OF GENES ARE PRESENTED IN EXTRA COLUMNS
            l_short = " ".join(l.split())
            l_list = l_short.split(" ")

            gene_name_list.append(l_list[0].strip(";"))
            gene_sgd_list.append(l_list[0].strip(";"))

            gene_oln = l_list[1 + extra_columns].strip(";")  # GET THE OLN NAME
            if (
                gene_oln == "GAG" or gene_oln == "POL"
            ):  # CHECK WHETHER THE OLN IS 'GAG' OR 'POL'. IF YES, TAKE THE NEXT COLUMN
                gene_name_list.append(l_list[2 + extra_columns].strip(";"))
                gene_oln_list.append(l_list[2 + extra_columns].strip(";"))
            else:
                gene_name_list.append(gene_oln)
                gene_oln_list.append(gene_oln)

            if (
                l_list[1 + extra_columns] == "GAG" or l_list[1 + extra_columns] == "POL"
            ):  # THESE ARE SEQUENCES THAT SOMETIMES OCCUR WHICH HAVE TO BE IGNORED.
                extra_columns = extra_columns + 1
            if extra_columns > 0:
                for n in range(extra_columns):
                    gene_name = l_list[1 + n].strip(";")
                    if not gene_name == "GAG" and not gene_name == "POL":
                        gene_name_list.append(gene_name)
            gene_counter += 1

    ###SAVING OLN LIST
    #    from datetime import date
    #    current_date = date.today()

    #    oln_saving_file = r"S_Cerevisiae_protein_oln_name_full_genome.txt"
    #    gene_oln_list_sorted = sorted(gene_oln_list)
    #    with open(oln_saving_file, 'w') as f:
    #
    ##        f.write("org=S. Cerevisiae ; type=Genomic ; naming='oln' ; source file='Yeast_Protein_Names.txt' ; creation date=%s using 'gene_names.py'\n" % current_date)
    #        for oln_name in gene_oln_list_sorted:
    #            f.write("%s\n" % oln_name)

    #    sgd_saving_file = r"S_Cerevisiae_protein_designation_name_full_genome.txt"
    #    gene_sgd_list_sorted = sorted(gene_sgd_list)
    #    with open(sgd_saving_file, 'w') as f:
    #
    #        f.write("org=S. Cerevisiae ; type=Genomic ; naming='designation' ; source file='Yeast_Protein_Names.txt' ; creation date=%s using 'gene_names.py'\n" % current_date)
    #        for sgd_name in gene_sgd_list_sorted:
    #            f.write("%s \n" % sgd_name)

    ###
    print("Number of genes found in file = ", gene_counter)
    return gene_name_list

